from __future__ import annotations

from dataclasses import dataclass, fields
from functools import singledispatch
from typing import TYPE_CHECKING, Literal, get_origin, get_type_hints

# note that types like Spanners are only used in annotations for concretes of the
# resolve generic, but we need to import at runtime, due to singledispatch looking
# up annotations
from ._gt_data import GTData, Spanners, ColInfoTypeEnum
from ._tbl_data import eval_select


if TYPE_CHECKING:
    from ._gt_data import TblData


# Locations ============================================================================
# TODO: these are called cells_* in gt. I prefixed them with Loc just to keep things
# straight while going through helpers.R, but no strong opinion on naming!


@dataclass
class Loc:
    """A location."""


@dataclass
class LocTitle(Loc):
    """A location for targeting the table title and subtitle."""

    groups: Literal["title", "subtitle"]


@dataclass
class LocStubhead(Loc):
    groups: Literal["stubhead"] = "stubhead"


@dataclass
class LocColumnSpanners(Loc):
    """A location for column spanners."""

    # TODO: these can also be tidy selectors
    ids: list[str]


@dataclass
class LocColumnLabels(Loc):
    # TODO: these can be tidyselectors
    columns: list[str]


@dataclass
class LocRowGroups(Loc):
    # TODO: these can be tidyselectors
    groups: list[str]


@dataclass
class LocStub(Loc):
    # TODO: these can be tidyselectors
    # TODO: can this take integers?
    rows: list[str]


@dataclass
class LocBody(Loc):
    # TODO: these can be tidyselectors
    columns: list[str]
    rows: list[str]


@dataclass
class LocSummary(Loc):
    # TODO: these can be tidyselectors
    groups: list[str]
    columns: list[str]
    rows: list[str]


@dataclass
class LocGrandSummary(Loc):
    # TODO: these can be tidyselectors
    columns: list[str]
    rows: list[str]


@dataclass
class LocStubSummary(Loc):
    # TODO: these can be tidyselectors
    groups: list[str]
    rows: list[str]


@dataclass
class LocStubGrandSummary(Loc):
    rows: list[str]


@dataclass
class LocFootnotes(Loc):
    groups: Literal["footnotes"] = "footnotes"


@dataclass
class LocSourceNotes(Loc):
    # This dataclass in R has a `groups` field, which is a literal value.
    # In python, we can use an isinstance check to determine we're seeing an
    # instance of this class
    groups: Literal["source_notes"] = "source_notes"


# Utils ================================================================================


def resolve_vector_i(expr: list[str], candidates: list[str], item_label: str) -> list[int]:
    """Return list of indices for candidates, selected by expr."""

    mask = resolve_vector_l(expr, candidates, item_label)
    return [ii for ii, val in enumerate(mask) if val]


def resolve_vector_l(expr: list[str], candidates: list[str], item_label: str) -> list[bool]:
    """Return list of logical index for candidates, selected by expr."""

    if not (isinstance(expr, list) and all(isinstance(x, str) for x in expr)):
        raise NotImplementedError("Selecting entries currently requires a list of strings.")

    set_expr = set(expr)
    if set_expr - set(candidates):
        missing = set_expr - set(candidates)
        raise ValueError(f"Cannot find these entries: {missing}")

    return [candidate in set_expr for candidate in candidates]


def resolve_cols_c(
    expr: list[str],
    data: GTData,
    strict: bool = True,
    excl_stub: bool = True,
    excl_group: bool = True,
    null_means: Literal["everything", "nothing"] = "everything",
) -> list[str]:
    selected = resolve_cols_i(expr, data, strict, excl_stub, excl_group, null_means)
    return [name_pos[0] for name_pos in selected]


def resolve_cols_i(
    expr: list[str],
    data: GTData,
    strict: bool = True,
    excl_stub: bool = True,
    excl_group: bool = True,
    null_means: Literal["everything", "nothing"] = "everything",
) -> list[tuple[str, int]]:
    """Return a tuple of (column name, position) pairs, selected by expr."""

    if isinstance(data, GTData):
        stub_var = data._boxhead.vars_from_type(ColInfoTypeEnum.stub)

        # TODO: special handling of "stub()"
        if "stub()" in expr:
            if len(stub_var):
                return [(stub_var[0], 1)]

            return []

        if not excl_stub:
            # In most cases we would want to exclude the column that
            # represents the stub but that isn't always the case (e.g.,
            # when considering the stub for column sizing); the `excl_stub`
            # argument will determine whether the stub column is obtained
            # for exclusion or not (if FALSE, we get NULL which removes the
            # stub, if present, from `cols_excl`)
            stub_var = None

        if not excl_group:
            # The columns that represent the group rows are usually
            # always excluded but in certain cases (i.e., `rows_add()`)
            # we may want to include this column
            _group_vars = data._boxhead.vars_from_type(ColInfoTypeEnum.row_group)
            group_var = _group_vars[0] if len(_group_vars) else None
        else:
            group_var = None

        cols_excl = [stub_var, group_var]

        tbl_data = data._tbl_data
    else:
        # TODO: is this path used? In the R program, cols_excl isn't set, so it seems
        # like it must not get used.
        tbl_data = data._tbl_data
        cols_excl = []

    selected = eval_select(tbl_data, expr, strict)
    return [name_pos for name_pos in selected if name_pos[0] not in cols_excl]


# Resolve generic ======================================================================


@singledispatch
def resolve(loc: Loc, *args, **kwargs) -> Loc:
    """Return a copy of location with lookups resolved (e.g. tidyselect on columns)."""
    raise NotImplementedError(f"Unsupported location type: {type(loc)}")


@resolve.register
def _(loc: LocColumnSpanners, spanners: Spanners) -> LocColumnSpanners:
    # unique labels (with order preserved)
    spanner_ids = [span.spanner_id for span in spanners]

    resolved_spanners_idx = resolve_vector_i(loc.ids, spanner_ids, item_label="spanner")
    resolved_spanners = [spanner_ids[idx] for idx in resolved_spanners_idx]

    # Create a list object
    return LocColumnSpanners(ids=resolved_spanners)



from __future__ import annotations
from typing import Optional, Union, List, Any, cast
from great_tables import _utils

default_fonts_list = [
    "-apple-system",
    "BlinkMacSystemFont",
    "Segoe UI",
    "Roboto",
    "Oxygen",
    "Ubuntu",
    "Cantarell",
    "Helvetica Neue",
    "Fira Sans",
    "Droid Sans",
    "Arial",
    "sans-serif",
]


class OptionsInfo:
    parameter: str
    scss: Optional[bool]
    category: Optional[str]
    type: Optional[str]
    value: Optional[Union[Any, List[str]]]

    def __init__(
        self,
        parameter: str,
        scss: Optional[bool] = None,
        category: Optional[str] = None,
        type: Optional[str] = None,
        value: Optional[Union[Any, List[str]]] = None,
    ):
        self.parameter = parameter
        self.scss = scss
        self.category = category
        self.type = type
        self.value = value


# fmt: off
class Options:
    def __init__(self):
        self._options: dict[str, OptionsInfo] = dict(
           (v.parameter, v) for v in [
            #           parameter                            scss    category            type        value
            OptionsInfo("container_width",                   False,  "container",        "px",       "auto"),
            OptionsInfo("container_height",                  False,  "container",        "px",       "auto"),
            OptionsInfo("container_overflow_x",              False,  "container",        "overflow", "auto"),
            OptionsInfo("container_overflow_y",              False,  "container",        "overflow", "auto"),
            OptionsInfo("table_id",                          False,  "table",            "value",    None),
            OptionsInfo("table_caption",                     False,  "table",            "value",    None),
            OptionsInfo("table_width",                        True,  "table",            "px",       "auto"),
            OptionsInfo("table_layout",                       True,  "table",            "value",    "fixed"),
            OptionsInfo("table_margin_left",                  True,  "table",            "px",       "auto"),
            OptionsInfo("table_margin_right",                 True,  "table",            "px",       "auto"),
            OptionsInfo("table_background_color",             True,  "table",            "value",    "#FFFFFF"),
            OptionsInfo("table_additional_css",              False,  "table",            "values",   None),
            OptionsInfo("table_font_names",                  False,  "table",            "values",   default_fonts_list),
            OptionsInfo("table_font_size",                    True,  "table",            "px",       "16px"),
            OptionsInfo("table_font_weight",                  True,  "table",            "value",    "normal"),
            OptionsInfo("table_font_style",                   True,  "table",            "value",    "normal"),
            OptionsInfo("table_font_color",                   True,  "table",            "value",    "#333333"),
            OptionsInfo("table_font_color_light",             True,  "table",            "value",    "#FFFFFF"),
            OptionsInfo("table_border_top_include",          False,  "table",            "boolean",  True),
            OptionsInfo("table_border_top_style",             True,  "table",            "value",    "solid"),
            OptionsInfo("table_border_top_width",             True,  "table",            "px",       "2px"),
            OptionsInfo("table_border_top_color",             True,  "table",            "value",    "#A8A8A8"),
            OptionsInfo("table_border_right_style",           True,  "table",            "value",    "none"),
            OptionsInfo("table_border_right_width",           True,  "table",            "px",       "2px"),
            OptionsInfo("table_border_right_color",           True,  "table",            "value",    "#D3D3D3"),
            OptionsInfo("table_border_bottom_include",       False,  "table",            "boolean",  True),
            OptionsInfo("table_border_bottom_style",          True,  "table",            "value",    "solid"),
            OptionsInfo("table_border_bottom_width",          True,  "table",            "px",       "2px"),
            OptionsInfo("table_border_bottom_color",          True,  "table",            "value",    "#A8A8A8"),
            OptionsInfo("table_border_left_style",            True,  "table",            "value",    "none"),
            OptionsInfo("table_border_left_width",            True,  "table",            "px",       "2px"),
            OptionsInfo("table_border_left_color",            True,  "table",            "value",    "#D3D3D3"),
            OptionsInfo("heading_background_color",           True,  "heading",          "value",    None),
            OptionsInfo("heading_align",                      True,  "heading",          "value",    "center"),
            OptionsInfo("heading_title_font_size",            True,  "heading",          "px",       "125%"),
            OptionsInfo("heading_title_font_weight",          True,  "heading",          "value",    "initial"),
            OptionsInfo("heading_subtitle_font_size",         True,  "heading",          "px",       "85%"),
            OptionsInfo("heading_subtitle_font_weight",       True,  "heading",          "value",    "initial"),
            OptionsInfo("heading_padding",                    True,  "heading",          "px",       "4px"),
            OptionsInfo("heading_padding_horizontal",         True,  "heading",          "px",       "5px"),
            OptionsInfo("heading_border_bottom_style",        True,  "heading",          "value",    "solid"),
            OptionsInfo("heading_border_bottom_width",        True,  "heading",          "px",       "2px"),
            OptionsInfo("heading_border_bottom_color",        True,  "heading",          "value",    "#D3D3D3"),
            OptionsInfo("heading_border_lr_style",            True,  "heading",          "value",    "none"),
            OptionsInfo("heading_border_lr_width",            True,  "heading",          "px",       "1px"),
            OptionsInfo("heading_border_lr_color",            True,  "heading",          "value",    "#D3D3D3"),
            OptionsInfo("column_labels_background_color",     True,  "column_labels",    "value",    None),
            OptionsInfo("column_labels_font_size",            True,  "column_labels",    "px",       "100%"),
            OptionsInfo("column_labels_font_weight",          True,  "column_labels",    "value",    "normal"),
            OptionsInfo("column_labels_text_transform",       True,  "column_labels",    "value",    "inherit"),
            OptionsInfo("column_labels_padding",              True,  "column_labels",    "px",       "5px"),
            OptionsInfo("column_labels_padding_horizontal",   True,  "column_labels",    "px",       "5px"),
            OptionsInfo("column_labels_vlines_style",         True,  "table_body",       "value",    "none"),
            OptionsInfo("column_labels_vlines_width",         True,  "table_body",       "px",       "1px"),
            OptionsInfo("column_labels_vlines_color",         True,  "table_body",       "value",    "#D3D3D3"),
            OptionsInfo("column_labels_border_top_style",     True,  "column_labels",    "value",    "solid"),
            OptionsInfo("column_labels_border_top_width",     True,  "column_labels",    "px",       "2px"),
            OptionsInfo("column_labels_border_top_color",     True,  "column_labels",    "value",    "#D3D3D3"),
            OptionsInfo("column_labels_border_bottom_style",  True,  "column_labels",    "value",    "solid"),
            OptionsInfo("column_labels_border_bottom_width",  True,  "column_labels",    "px",       "2px"),
            OptionsInfo("column_labels_border_bottom_color",  True,  "column_labels",    "value",    "#D3D3D3"),
            OptionsInfo("column_labels_border_lr_style",      True,  "column_labels",    "value",    "none"),
            OptionsInfo("column_labels_border_lr_width",      True,  "column_labels",    "px",       "1px"),
            OptionsInfo("column_labels_border_lr_color",      True,  "column_labels",    "value",    "#D3D3D3"),
            OptionsInfo("column_labels_hidden",              False,  "column_labels",    "boolean",  False),
            OptionsInfo("row_group_background_color",         True,  "row_group",        "value",    None),
            OptionsInfo("row_group_font_size",                True,  "row_group",        "px",       "100%"),
            OptionsInfo("row_group_font_weight",              True,  "row_group",        "value",    "initial"),
            OptionsInfo("row_group_text_transform",           True,  "row_group",        "value",    "inherit"),
            OptionsInfo("row_group_padding",                  True,  "row_group",        "px",       "8px"),
            OptionsInfo("row_group_padding_horizontal",       True,  "row_group",        "px",       "5px"),
            OptionsInfo("row_group_border_top_style",         True,  "row_group",        "value",    "solid"),
            OptionsInfo("row_group_border_top_width",         True,  "row_group",        "px",       "2px"),
            OptionsInfo("row_group_border_top_color",         True,  "row_group",        "value",    "#D3D3D3"),
            OptionsInfo("row_group_border_right_style",       True,  "row_group",        "value",    "none"),
            OptionsInfo("row_group_border_right_width",       True,  "row_group",        "px",       "1px"),
            OptionsInfo("row_group_border_right_color",       True,  "row_group",        "value",    "#D3D3D3"),
            OptionsInfo("row_group_border_bottom_style",      True,  "row_group",        "value",    "solid"),
            OptionsInfo("row_group_border_bottom_width",      True,  "row_group",        "px",       "2px"),
            OptionsInfo("row_group_border_bottom_color",      True,  "row_group",        "value",    "#D3D3D3"),
            OptionsInfo("row_group_border_left_style",        True,  "row_group",        "value",    "none"),
            OptionsInfo("row_group_border_left_width",        True,  "row_group",        "px",       "1px"),
            OptionsInfo("row_group_border_left_color",        True,  "row_group",        "value",    "#D3D3D3"),
            OptionsInfo("row_group_default_label",           False,  "row_group",        "value",    None),
            OptionsInfo("row_group_as_column",               False,  "row_group",        "boolean",  False),
            OptionsInfo("table_body_hlines_style",            True,  "table_body",       "value",    "solid"),
            OptionsInfo("table_body_hlines_width",            True,  "table_body",       "px",       "1px"),
            OptionsInfo("table_body_hlines_color",            True,  "table_body",       "value",    "#D3D3D3"),
            OptionsInfo("table_body_vlines_style",            True,  "table_body",       "value",    "none"),
            OptionsInfo("table_body_vlines_width",            True,  "table_body",       "px",       "1px"),
            OptionsInfo("table_body_vlines_color",            True,  "table_body",       "value",    "#D3D3D3"),
            OptionsInfo("table_body_border_top_style",        True,  "table_body",       "value",    "solid"),
            OptionsInfo("table_body_border_top_width",        True,  "table_body",       "px",       "2px"),
            OptionsInfo("table_body_border_top_color",        True,  "table_body",       "value",    "#D3D3D3"),
            OptionsInfo("table_body_border_bottom_style",     True,  "table_body",       "value",    "solid"),
            OptionsInfo("table_body_border_bottom_width",     True,  "table_body",       "px",       "2px"),
            OptionsInfo("table_body_border_bottom_color",     True,  "table_body",       "value",    "#D3D3D3"),
            OptionsInfo("data_row_padding",                   True,  "data_row",         "px",       "8px"),
            OptionsInfo("data_row_padding_horizontal",        True,  "data_row",         "px",       "5px"),
            OptionsInfo("stub_background_color",              True,  "stub",             "value",    None),
            OptionsInfo("stub_font_size",                     True,  "stub",             "px",       "100%"),
            OptionsInfo("stub_font_weight",                   True,  "stub",             "value",    "initial"),
            OptionsInfo("stub_text_transform",                True,  "stub",             "value",    "inherit"),
            OptionsInfo("stub_border_style",                  True,  "stub",             "value",    "solid"),
            OptionsInfo("stub_border_width",                  True,  "stub",             "px",       "2px"),
            OptionsInfo("stub_border_color",                  True,  "stub",             "value",    "#D3D3D3"),
            OptionsInfo("stub_row_group_background_color",    True,  "stub",             "value",    None),
            OptionsInfo("stub_row_group_font_size",           True,  "stub",             "px",       "100%"),
            OptionsInfo("stub_row_group_font_weight",         True,  "stub",             "value",    "initial"),
            OptionsInfo("stub_row_group_text_transform",      True,  "stub",             "value",    "inherit"),
            OptionsInfo("stub_row_group_border_style",        True,  "stub",             "value",    "solid"),
            OptionsInfo("stub_row_group_border_width",        True,  "stub",             "px",       "2px"),
            OptionsInfo("stub_row_group_border_color",        True,  "stub",             "value",    "#D3D3D3"),
            OptionsInfo("summary_row_padding",                True,  "summary_row",      "px",       "8px"),
            OptionsInfo("summary_row_padding_horizontal",     True,  "summary_row",      "px",       "5px"),
            OptionsInfo("summary_row_background_color",       True,  "summary_row",      "value",    None),
            OptionsInfo("summary_row_text_transform",         True,  "summary_row",      "value",    "inherit"),
            OptionsInfo("summary_row_border_style",           True,  "summary_row",      "value",    "solid"),
            OptionsInfo("summary_row_border_width",           True,  "summary_row",      "px",       "2px"),
            OptionsInfo("summary_row_border_color",           True,  "summary_row",      "value",    "#D3D3D3"),
            OptionsInfo("grand_summary_row_padding",          True,  "grand_summary_row", "px",      "8px"),
            OptionsInfo("grand_summary_row_padding_horizontal",True, "grand_summary_row", "px",      "5px"),
            OptionsInfo("grand_summary_row_background_color", True,  "grand_summary_row", "value",   None),
            OptionsInfo("grand_summary_row_text_transform",   True,  "grand_summary_row", "value",   "inherit"),
            OptionsInfo("grand_summary_row_border_style",     True,  "grand_summary_row", "value",   "double"),
            OptionsInfo("grand_summary_row_border_width",     True,  "grand_summary_row", "px",      "6px"),
            OptionsInfo("grand_summary_row_border_color",     True,  "grand_summary_row", "value",   "#D3D3D3"),
            OptionsInfo("footnotes_font_size",                True,  "footnotes",        "px",       "90%"),
            OptionsInfo("footnotes_padding",                  True,  "footnotes",        "px",       "4px"),
            OptionsInfo("footnotes_padding_horizontal",       True,  "footnotes",        "px",       "5px"),
            OptionsInfo("footnotes_background_color",         True,  "footnotes",        "value",    None),
            OptionsInfo("footnotes_margin",                   True,  "footnotes",        "px",       "0px"),
            OptionsInfo("footnotes_border_bottom_style",      True,  "footnotes",        "value",    "none"),
            OptionsInfo("footnotes_border_bottom_width",      True,  "footnotes",        "px",       "2px"),
            OptionsInfo("footnotes_border_bottom_color",      True,  "footnotes",        "value",    "#D3D3D3"),
            OptionsInfo("footnotes_border_lr_style",          True,  "footnotes",        "value",    "none"),
            OptionsInfo("footnotes_border_lr_width",          True,  "footnotes",        "px",       "2px"),
            OptionsInfo("footnotes_border_lr_color",          True,  "footnotes",        "value",    "#D3D3D3"),
            OptionsInfo("footnotes_marks" ,                  False,  "footnotes",        "values",   "numbers"),
            OptionsInfo("footnotes_multiline",               False,  "footnotes",        "boolean",  True),
            OptionsInfo("footnotes_sep",                     False,  "footnotes",        "value",    " "),
            OptionsInfo("source_notes_padding",               True,  "source_notes",     "px",       "4px"),
            OptionsInfo("source_notes_padding_horizontal",    True,  "source_notes",     "px",       "5px"),
            OptionsInfo("source_notes_background_color",      True,  "source_notes",     "value",    None),
            OptionsInfo("source_notes_font_size",             True,  "source_notes",     "px",       "90%"),
            OptionsInfo("source_notes_border_bottom_style",   True,  "source_notes",     "value",    "none"),
            OptionsInfo("source_notes_border_bottom_width",   True,  "source_notes",     "px",       "2px"),
            OptionsInfo("source_notes_border_bottom_color",   True,  "source_notes",     "value",    "#D3D3D3"),
            OptionsInfo("source_notes_border_lr_style",       True,  "source_notes",     "value",    "none"),
            OptionsInfo("source_notes_border_lr_width",       True,  "source_notes",     "px",       "2px"),
            OptionsInfo("source_notes_border_lr_color",       True,  "source_notes",     "value",    "#D3D3D3"),
            OptionsInfo("source_notes_multiline",            False,  "source_notes",     "boolean",  True),
            OptionsInfo("source_notes_sep",                  False,  "source_notes",     "value",    " "),
            OptionsInfo("row_striping_background_color",      True,  "row",              "value",    "rgba(128,128,128,0.05)"),
            OptionsInfo("row_striping_include_stub",         False,  "row",              "boolean",  False),
            OptionsInfo("row_striping_include_table_body",   False,  "row",              "boolean",  False),
            OptionsInfo("page_orientation",                  False,  "page",             "value",    "portrait"),
            OptionsInfo("page_numbering",                    False,  "page",             "boolean",  False),
            OptionsInfo("page_header_use_tbl_headings",      False,  "page",             "boolean",  False),
            OptionsInfo("page_footer_use_tbl_notes",         False,  "page",             "boolean",  False),
            OptionsInfo("page_width",                        False,  "page",             "value",    "8.5in"),
            OptionsInfo("page_height",                       False,  "page",             "value",    "11.0in"),
            OptionsInfo("page_margin_left",                  False,  "page",             "value",    "1.0in"),
            OptionsInfo("page_margin_right",                 False,  "page",             "value",    "1.0in"),
            OptionsInfo("page_margin_top",                   False,  "page",             "value",    "1.0in"),
            OptionsInfo("page_margin_bottom",                False,  "page",             "value",    "1.0in"),
            OptionsInfo("page_header_height",                False,  "page",             "value",    "0.5in"),
            OptionsInfo("page_footer_height",                False,  "page",             "value",    "0.5in"),
        ])
# fmt: on

    def _get_all_options_keys(self) -> List[Union[str, None]]:
        return [x.parameter for x in self._options.values()]

    #def __setattr__(self, __name: str, __value: Any) -> None:
    #    self._options[__name].value = __value

    def __getattr__(self, __name: str) -> OptionsInfo:
        return self._options[__name]
        # use this like
        #
        # options = Options()
        # options.foo.value = "bar"
        # print(options.foo.type)

    #def _get_option_type(self, option: str) -> Union[Any, List[str]]:
    #    return self._options[option].type

    def _get_option_value(self, option: str) -> Union[Any, List[str]]:
        return self._options[option].value

    def _set_option_value(self, option: str, value: Any):
       self._options[option].value = value
       return self


class OptionsAPI:
    def tab_options(
            self,
            container_width: Optional[str] = None,
            container_height: Optional[str] = None,
            container_overflow_x: Optional[str] = None,
            container_overflow_y: Optional[str] = None,
            table_width: Optional[str] = None,
            table_layout: Optional[str] = None,
            table_align: Optional[str] = None,
            table_margin_left: Optional[str] = None,
            table_margin_right: Optional[str] = None,
            table_background_color: Optional[str] = None,
            table_additional_css: Optional[str] = None,
            table_font_names: Optional[str] = None,
            table_font_size: Optional[str] = None,
            table_font_weight: Optional[str] = None,
            table_font_style: Optional[str] = None,
            table_font_color: Optional[str] = None,
            table_font_color_light: Optional[str] = None,
            table_border_top_style: Optional[str] = None,
            table_border_top_width: Optional[str] = None,
            table_border_top_color: Optional[str] = None,
            table_border_right_style: Optional[str] = None,
            table_border_right_width: Optional[str] = None,
            table_border_right_color: Optional[str] = None,
            table_border_bottom_style: Optional[str] = None,
            table_border_bottom_width: Optional[str] = None,
            table_border_bottom_color: Optional[str] = None,
            table_border_left_style: Optional[str] = None,
            table_border_left_width: Optional[str] = None,
            table_border_left_color: Optional[str] = None,
            heading_background_color: Optional[str] = None,
            heading_align: Optional[str] = None,
            heading_title_font_size: Optional[str] = None,
            heading_title_font_weight: Optional[str] = None,
            heading_subtitle_font_size: Optional[str] = None,
            heading_subtitle_font_weight: Optional[str] = None,
            heading_padding: Optional[str] = None,
            heading_padding_horizontal: Optional[str] = None,
            heading_border_bottom_style: Optional[str] = None,
            heading_border_bottom_width: Optional[str] = None,
            heading_border_bottom_color: Optional[str] = None,
            heading_border_lr_style: Optional[str] = None,
            heading_border_lr_width: Optional[str] = None,
            heading_border_lr_color: Optional[str] = None,
            column_labels_background_color: Optional[str] = None,
            column_labels_font_size: Optional[str] = None,
            column_labels_font_weight: Optional[str] = None,
            column_labels_text_transform: Optional[str] = None,
            column_labels_padding: Optional[str] = None,
            column_labels_padding_horizontal: Optional[str] = None,
            column_labels_vlines_style: Optional[str] = None,
            column_labels_vlines_width: Optional[str] = None,
            column_labels_vlines_color: Optional[str] = None,
            column_labels_border_top_style: Optional[str] = None,
            column_labels_border_top_width: Optional[str] = None,
            column_labels_border_top_color: Optional[str] = None,
            column_labels_border_bottom_style: Optional[str] = None,
            column_labels_border_bottom_width: Optional[str] = None,
            column_labels_border_bottom_color: Optional[str] = None,
            column_labels_border_lr_style: Optional[str] = None,
            column_labels_border_lr_width: Optional[str] = None,
            column_labels_border_lr_color: Optional[str] = None,
            column_labels_hidden: Optional[bool] = None,
            row_group_background_color: Optional[str] = None,
            row_group_font_size: Optional[str] = None,
            row_group_font_weight: Optional[str] = None,
            row_group_text_transform: Optional[str] = None,
            row_group_padding: Optional[str] = None,
            row_group_padding_horizontal: Optional[str] = None,
            row_group_border_top_style: Optional[str] = None,
            row_group_border_top_width: Optional[str] = None,
            row_group_border_top_color: Optional[str] = None,
            row_group_border_bottom_style: Optional[str] = None,
            row_group_border_bottom_width: Optional[str] = None,
            row_group_border_bottom_color: Optional[str] = None,
            row_group_border_left_style: Optional[str] = None,
            row_group_border_left_width: Optional[str] = None,
            row_group_border_left_color: Optional[str] = None,
            row_group_border_right_style: Optional[str] = None,
            row_group_border_right_width: Optional[str] = None,
            row_group_border_right_color: Optional[str] = None,
            row_group_default_label: Optional[str] = None,
            row_group_as_column: Optional[bool] = None,
            table_body_hlines_style: Optional[str] = None,
            table_body_hlines_width: Optional[str] = None,
            table_body_hlines_color: Optional[str] = None,
            table_body_vlines_style: Optional[str] = None,
            table_body_vlines_width: Optional[str] = None,
            table_body_vlines_color: Optional[str] = None,
            table_body_border_top_style: Optional[str] = None,
            table_body_border_top_width: Optional[str] = None,
            table_body_border_top_color: Optional[str] = None,
            table_body_border_bottom_style: Optional[str] = None,
            table_body_border_bottom_width: Optional[str] = None,
            table_body_border_bottom_color: Optional[str] = None,
            stub_background_color: Optional[str] = None,
            stub_font_size: Optional[str] = None,
            stub_font_weight: Optional[str] = None,
            stub_text_transform: Optional[str] = None,
            stub_border_style: Optional[str] = None,
            stub_border_width: Optional[str] = None,
            stub_border_color: Optional[str] = None,
            stub_row_group_font_size: Optional[str] = None,
            stub_row_group_font_weight: Optional[str] = None,
            stub_row_group_text_transform: Optional[str] = None,
            stub_row_group_border_style: Optional[str] = None,
            stub_row_group_border_width: Optional[str] = None,
            stub_row_group_border_color: Optional[str] = None,
            data_row_padding: Optional[str] = None,
            data_row_padding_horizontal: Optional[str] = None,
            summary_row_background_color: Optional[str] = None,
            summary_row_text_transform: Optional[str] = None,
            summary_row_padding: Optional[str] = None,
            summary_row_padding_horizontal: Optional[str] = None,
            summary_row_border_style: Optional[str] = None,
            summary_row_border_width: Optional[str] = None,
            summary_row_border_color: Optional[str] = None,
            grand_summary_row_background_color: Optional[str] = None,
            grand_summary_row_text_transform: Optional[str] = None,
            grand_summary_row_padding: Optional[str] = None,
            grand_summary_row_padding_horizontal: Optional[str] = None,
            grand_summary_row_border_style: Optional[str] = None,
            grand_summary_row_border_width: Optional[str] = None,
            grand_summary_row_border_color: Optional[str] = None,
            footnotes_background_color: Optional[str] = None,
            footnotes_font_size: Optional[str] = None,
            footnotes_padding: Optional[str] = None,
            footnotes_padding_horizontal: Optional[str] = None,
            footnotes_border_bottom_style: Optional[str] = None,
            footnotes_border_bottom_width: Optional[str] = None,
            footnotes_border_bottom_color: Optional[str] = None,
            footnotes_border_lr_style: Optional[str] = None,
            footnotes_border_lr_width: Optional[str] = None,
            footnotes_border_lr_color: Optional[str] = None,
            footnotes_marks: Optional[Union[str, List[str]]] = None,
            footnotes_multiline: Optional[bool] = None,
            footnotes_sep: Optional[str] = None,
            source_notes_background_color: Optional[str] = None,
            source_notes_font_size: Optional[str] = None,
            source_notes_padding: Optional[str] = None,
            source_notes_padding_horizontal: Optional[str] = None,
            source_notes_border_bottom_style: Optional[str] = None,
            source_notes_border_bottom_width: Optional[str] = None,
            source_notes_border_bottom_color: Optional[str] = None,
            source_notes_border_lr_style: Optional[str] = None,
            source_notes_border_lr_width: Optional[str] = None,
            source_notes_border_lr_color: Optional[str] = None,
            source_notes_multiline: Optional[bool] = None,
            source_notes_sep: Optional[str] = None,
            row_striping_background_color: Optional[str] = None,
            row_striping_include_stub: Optional[bool] = None,
            row_striping_include_table_body: Optional[bool] = None,
            page_orientation: Optional[str] = None,
            page_numbering: Optional[bool] = None,
            page_header_use_tbl_headings: Optional[bool] = None,
            page_footer_use_tbl_notes: Optional[bool] = None,
            page_width: Optional[str] = None,
            page_height: Optional[str] = None,
            page_margin_left: Optional[str] = None,
            page_margin_right: Optional[str] = None,
            page_margin_top: Optional[str] = None,
            page_margin_bottom: Optional[str] = None,
            page_header_height: Optional[str] = None,
            page_footer_height: Optional[str] = None,
    ):
        """
        Modify the table output options

        Modify the options available in a table. These options are named by the
        components, the subcomponents, and the element that can adjusted.

        Parameters
        ----------
        container_width,container_height
            The width and height of the table's container. Can be specified as
            a single-length character with units of pixels or as a percentage.
            If provided as a scalar numeric value, it is assumed that the
            value is given in units of pixels. The `px()` and `pct()` helper
            functions can also be used to pass in numeric values and obtain values
            as pixel or percent units.

        container_overflow_x,container_overflow_y
            Options to enable scrolling in the horizontal and vertical directions
            when the table content overflows the container dimensions. Using `True`
            (the default for both) means that horizontal or vertical scrolling is
            enabled to view the entire table in those directions. With `False`, the
            table may be clipped if the table width or height exceeds the
            `container_width` or `container_height`.

         table_width
            The width of the table. Can be specified as a single-length character
            with units of pixels or as a percentage. If provided as a scalar
            numeric value, it is assumed that the value is given in units of pixels.
            The `px()` and `pct()` helper functions can also be used to pass in
            numeric values and obtain values as pixel or percent units.

         table_layout
            The value for the `table-layout` CSS style in the HTML output context.
            By default, this is `"fixed"` but another valid option is `"auto"`.

         table_align
            The horizontal alignment of the table in its container. By default, this
            is `"center"`. Other options are `"left"` and `"right"`. This will
            automatically set `table_margin_left` and `table_margin_right` to the
            appropriate values.

         table_margin_left,table_margin_right
            The size of the margins on the left and right of the table within the
            container. Can be specified as a single-length character with units of
            pixels or as a percentage. If provided as a scalar numeric value, it is
            assumed that the value is given in units of pixels. The `px()` and
            `pct()` helper functions can also be used to pass in numeric values and
            obtain values as pixel or percent units. Using `table_margin_left` or
            `table_margin_right` will overwrite any values set by `table_align`.

         table_background_color,heading_background_color,column_labels_background_color,row_group_background_color,stub_background_color,summary_row_background_color,grand_summary_row_background_color,footnotes_background_color,source_notes_background_color
           Background colors for the parent element `table` and the following child
           elements: `heading`, `column_labels`, `row_group`, `stub`, `summary_row`,
           `grand_summary_row`, `footnotes`, and `source_notes`. A color name or a
           hexadecimal color code should be provided.

         table_additional_css
            This option can be used to supply an additional block of CSS rules to be
            applied after the automatically generated table CSS.

         table_font_names
            The names of the fonts used for the table. This should be provided as a
            list of font names. If the first font isn't available, then the next font
            is tried (and so on).

         table_font_style
            The font style for the table. Can be one of either `"normal"`, `"italic"`,
            or `"oblique"`.

         table_font_color,table_font_color_light
            The text color used throughout the table. There are two variants:
            `table_font_color` is for text overlaid on lighter background colors, and
            `table_font_color_light` is automatically used when text needs to be
            overlaid on darker background colors. A color name or a hexadecimal color
            code should be provided.

         table_font_size,heading_title_font_size,heading_subtitle_font_size,column_labels_font_size,row_group_font_size,stub_font_size,footnotes_font_size,source_notes_font_size
            The font sizes for the parent text element `table` and the following child
            elements: `heading_title`, `heading_subtitle`, `column_labels`,
            `row_group`, `footnotes`, and `source_notes`. Can be specified as a
            string with units of pixels (e.g., `"12px"`) or as a percentage (e.g.,
            `"80%"`). If provided as a scalar numeric value, it is assumed that the
            value is given in units of pixels. The `px()` and `pct()` helper
            functions can also be used to pass in numeric values and obtain values
            as pixel or percentage units.

         heading_align
            Controls the horizontal alignment of the heading title and subtitle. We can
            either use `"center"`, `"left"`, or `"right"`.

         table_font_weight,heading_title_font_weight,heading_subtitle_font_weight,column_labels_font_weight,row_group_font_weight,stub_font_weight
            The font weights of the table, `heading_title`, `heading_subtitle`,
            `column_labels`, `row_group`, and `stub` text elements. Can be a text-based
            keyword such as `"normal"`, `"bold"`, `"lighter"`, `"bolder"`, or, a
            numeric value between `1` and `1000`, inclusive. Note that only variable
            fonts may support the numeric mapping of weight.

         column_labels_text_transform,row_group_text_transform,stub_text_transform,summary_row_text_transform,grand_summary_row_text_transform
            Options to apply text transformations to the `column_labels`, `row_group`,
            `stub`, `summary_row`, and `grand_summary_row` text elements. Either of the
            `"uppercase"`, `"lowercase"`, or `"capitalize"` keywords can be used.

         heading_padding,column_labels_padding,data_row_padding,row_group_padding,summary_row_padding,grand_summary_row_padding,footnotes_padding,source_notes_padding
            The amount of vertical padding to incorporate in the `heading` (title and
            subtitle), the `column_labels` (this includes the column spanners), the row
            group labels (`row_group_padding`), in the body/stub rows
            (`data_row_padding`), in summary rows (`summary_row_padding` or
            `grand_summary_row_padding`), or in the footnotes and source notes
            (`footnotes_padding` and `source_notes_padding`).

         heading_padding_horizontal,column_labels_padding_horizontal,data_row_padding_horizontal,row_group_padding_horizontal,summary_row_padding_horizontal,grand_summary_row_padding_horizontal,footnotes_padding_horizontal,source_notes_padding_horizontal
            The amount of horizontal padding to incorporate in the `heading` (title and
            subtitle), the `column_labels` (this includes the column spanners), the row
            group labels (`row_group_padding_horizontal`), in the body/stub rows
            (`data_row_padding`), in summary rows (`summary_row_padding_horizontal` or
            `grand_summary_row_padding_horizontal`), or in the footnotes and source
            notes (`footnotes_padding_horizontal` and `source_notes_padding_horizontal`).

         table_border_top_style,table_border_top_width,table_border_top_color,table_border_right_style,table_border_right_width,table_border_right_color,table_border_bottom_style,table_border_bottom_width,table_border_bottom_color,table_border_left_style,table_border_left_width,table_border_left_color
            The style, width, and color properties of the table's absolute top and
            absolute bottom borders.

         heading_border_bottom_style,heading_border_bottom_width,heading_border_bottom_color
            The style, width, and color properties of the header's bottom border. This
            border shares space with that of the `column_labels` location. If the
            `width` of this border is larger, then it will be the visible border.

         heading_border_lr_style,heading_border_lr_width,heading_border_lr_color
            The style, width, and color properties for the left and right borders of
            the `heading` location.

         column_labels_vlines_style,column_labels_vlines_width,column_labels_vlines_color
            The style, width, and color properties for all vertical lines ('vlines')
            of the the `column_labels`.

         column_labels_border_top_style,column_labels_border_top_width,column_labels_border_top_color
            The style, width, and color properties for the top border of the
            `column_labels` location. This border shares space with that of the
            `heading` location. If the `width` of this border is larger, then it will
            be the visible border.

         column_labels_border_bottom_style,column_labels_border_bottom_width,column_labels_border_bottom_color
            The style, width, and color properties for the bottom border of the
            `column_labels` location.

         column_labels_border_lr_style,column_labels_border_lr_width,column_labels_border_lr_color
            The style, width, and color properties for the left and right borders of
            the `column_labels` location.

         column_labels_hidden
            An option to hide the column labels. If providing `TRUE` then the entire
            `column_labels` location won't be seen and the table header (if present)
            will collapse downward.

         row_group_border_top_style,row_group_border_top_width,row_group_border_top_color,row_group_border_bottom_style,row_group_border_bottom_width,row_group_border_bottom_color,row_group_border_left_style,row_group_border_left_width,row_group_border_left_color,row_group_border_right_style,row_group_border_right_width,row_group_border_right_color
            The style, width, and color properties for all top, bottom, left, and right
            borders of the `row_group` location.

         table_body_hlines_style,table_body_hlines_width,table_body_hlines_color,table_body_vlines_style,table_body_vlines_width,table_body_vlines_color
            The style, width, and color properties for all horizontal lines ('hlines')
            and vertical lines ('vlines') in the `table_body`.

         table_body_border_top_style,table_body_border_top_width,table_body_border_top_color,table_body_border_bottom_style,table_body_border_bottom_width,table_body_border_bottom_color
            The style, width, and color properties for all top and bottom borders of
            the `table_body` location.

         stub_border_style,stub_border_width,stub_border_color
            The style, width, and color properties for the vertical border of the table
            stub.

         stub_row_group_font_size,stub_row_group_font_weight,stub_row_group_text_transform,stub_row_group_border_style,stub_row_group_border_width,stub_row_group_border_color
            Options for the row group column in the stub (made possible when using
            `row_group_as_column = TRUE`). The defaults for these options mirror that
            of the `stub.*` variants (except for `stub_row_group_border_width`, which
            is `"1px"` instead of `"2px"`).

         row_group_default_label
            An option to set a default row group label for any rows not formally placed
            in a row group named by `group` in any call of `tab_row_group()`. If this
            is set as `None` and there are rows that haven't been placed into
            a row group (where one or more row groups already exist), those rows will
            be automatically placed into a row group without a label.

         row_group_as_column
            How should row groups be structured? By default, they are separate rows
            that lie above the each of the groups. Setting this to `True` will
            structure row group labels are columns to the far left of the table.

         summary_row_border_style,summary_row_border_width,summary_row_border_color
            The style, width, and color properties for all horizontal borders of the
            `summary_row` location.

         grand_summary_row_border_style,grand_summary_row_border_width,grand_summary_row_border_color
            The style, width, and color properties for the top borders of the
            `grand_summary_row` location.

         footnotes_border_bottom_style,footnotes_border_bottom_width,footnotes_border_bottom_color
            The style, width, and color properties for the bottom border of the
            `footnotes` location.

         footnotes_border_lr_style,footnotes_border_lr_width,footnotes_border_lr_color
            The style, width, and color properties for the left and right borders of
            the `footnotes` location.

         footnotes_marks
            The set of sequential marks used to reference and identify each of the
            footnotes (same input as the [opt_footnote_marks()] function. We can
            supply a list that represents the series of footnote marks. This
            list is recycled when its usage goes beyond the length of the set. At
            each cycle, the marks are simply combined (e.g., `*` -> `**` ->
            `***`). The option exists for providing keywords for certain types of
            footnote marks. The keyword `"numbers"` (the default, indicating that we
            want to use numeric marks). We can use lowercase `"letters"` or uppercase
            `"LETTERS"`. There is the option for using a traditional symbol set where
            `"standard"` provides four symbols, and, `"extended"` adds two more
            symbols, making six.

         footnotes_multiline,source_notes_multiline
            An option to either put footnotes and source notes in separate lines
            (the default, or `True`) or render them as a continuous line of text
            with `footnotes_sep` providing the separator (by default `" "`) between
            notes.

         footnotes_sep,source_notes_sep
            The separating characters between adjacent footnotes and source notes
            in their respective footer sections when rendered as a continuous
            line of text (when `footnotes_multiline is False`). The default value
            is a single space character (`" "`).

         source_notes_border_bottom_style,source_notes_border_bottom_width,source_notes_border_bottom_color
            The style, width, and color properties for the bottom border of the
            `source_notes` location.

         source_notes_border_lr_style,source_notes_border_lr_width,source_notes_border_lr_color
            The style, width, and color properties for the left and right borders of
            the `source_notes` location.

         row_striping_background_color
            The background color for striped table body rows. A color name or a
            hexadecimal color code should be provided.

         row_striping_include_stub
            An option for whether to include the stub when striping rows.

         row_striping_include_table_body
            An option for whether to include the table body when striping rows.

         page_orientation
            For RTF output, this provides an two options for page orientation:
            `"portrait"` (the default) and `"landscape"`.

         page_numbering
            Within RTF output, should page numbering be displayed? By default,
            this is set to `False` but if `True` then page numbering text will be
            added to the document header.

         page_header_use_tbl_headings
            If `True` then RTF output tables will migrate all table headings
            (including the table title and all column labels) to the page header.
            This page header content will repeat across pages. By default, this is
            `False`.

         page_footer_use_tbl_notes
            If `True` then RTF output tables will migrate all table footer content
            (this includes footnotes and source notes) to the page footer. This page
            footer content will repeat across pages. By default, this is `False`.

         page_width,page_height
            The page width and height in the standard portrait orientation. This is
            for RTF table output and the default values (in inches) are `8.5in` and
            `11.0in`.

         page_margin_left,page_margin_right,page_margin_top,page_margin_bottom
            For RTF table output, these options correspond to the left, right, top, and
            bottom page margins. The default values for each of these is `1.0in`.

         page_header_height,page_footer_height
            The heights of the page header and footer for RTF table outputs. Default
            values for both are `0.5in`.

        Returns
        -------

        GT
            Result of the table operation.
        """
        saved_args = locals()

        del saved_args["self"]

        modified_args = {k: v for k, v in saved_args.items() if v is not None}

        for i in range(len(modified_args)):
            self._options._set_option_value(
                option=list(modified_args.keys())[i],
                value=list(modified_args.values())[i]
            )

        return self


    def opt_footnote_marks(self, marks: Union[str, List[str]] = "numbers"):
        """
        Option to modify the set of footnote marks.

        Alter the footnote marks for any footnotes that may be present in the table.
        Either a list of marks can be provided (including Unicode characters), or,
        a specific keyword could be used to signify a preset sequence. This function
        serves as a shortcut for using `tab_options(footnotes_marks = <marks>)`

        We can supply a list of strings will represent the series of marks.
        The series of footnote marks is recycled when its usage goes beyond the
        length of the set. At each cycle, the marks are simply doubled, tripled, and
        so on (e.g., `*` -> `**` -> `***`). The option exists for providing keywords
        for certain types of footnote marks. The keywords are:

        - `"numbers"`: numeric marks, they begin from 1 and these marks are not
        subject to recycling behavior
        - `"letters"`: lowercase alphabetic marks. Same as using the `gt.letters()`
        function which produces a list of 26 lowercase letters from the Roman
        alphabet
        - `"LETTERS"`: uppercase alphabetic marks. Same as using the `gt.LETTERS()`
        function which produces a list of 26 uppercase letters from the Roman
        alphabet
        - `"standard"`: symbolic marks, four symbols in total
        - `"extended"`: symbolic marks, extends the standard set by adding two
        more symbols, making six

        Parameters
        ----------

        marks
            Either a list of strings that will represent the series of marks or a
            keyword string that represents a preset sequence of marks. The valid
            keywords are: `"numbers"` (for numeric marks), `"letters"` and `"LETTERS"`
            (for lowercase and uppercase alphabetic marks), `"standard"` (for a
            traditional set of four symbol marks), and `"extended"` (which adds two
            more symbols to the standard set).

        Returns
        -------

        GT
            Result of the table operation.
        """
        # Validate the marks keyword passed in as a string
        if marks is str:
            marks = _utils._match_arg(
                x=cast(str, marks),
                lst=["numbers", "letters", "LETTERS", "standard", "extended"],
            )

        self.tab_options(footnotes_marks=marks)
        return self


    def opt_row_striping(self, row_striping: bool = True):
        """
        Option to add or remove row striping.

        By default, a gt*table does not have row striping enabled. However, this
        method allows us to easily enable or disable striped rows in the table
        body. It's a convenient shortcut for
        `gt.tab_options(row_striping_include_table_body = <True|False>)`.

        Parameters
        ----------

        row_striping (bool):
            A boolean that indicates whether row striping should be added or
            removed. Defaults to True.

        Returns
        -------

        GT
            Result of the table operation.
        """
        self.tab_options(row_striping_include_table_body=row_striping)
        return self


    def opt_align_table_header(self, align: str = "center"):
        """
        Option to align the table header.

        By default, a table header added to a gt table has center alignment
        for both the title and the subtitle elements. This function allows us to
        easily set the horizontal alignment of the title and subtitle to the left
        or right by using the `"align"` argument. This function serves as a
        convenient shortcut for `gt.tab_options(heading.align = <align>)`.

        Parameters
        ----------

        align (str):
            The alignment of the title and subtitle elements in the table
            header. Options are `"center"` (the default), `"left"`, or
            `"right"`.

        Returns
        -------

        GT
            Result of the table operation.
        """

        align = _utils._match_arg(x=align, lst=["left", "center", "right"])

        self.tab_options(heading_align=align)
        return self

    # TODO: create the `opt_vertical_padding()` function

    # TODO: create the `opt_horizontal_padding()` function

    def opt_all_caps(
        self,
        all_caps: bool = True,
        locations: Union[str, List[str]] = ["column_labels", "stub", "row_group"],
    ):
        """
        Option to use all caps in select table locations.

        Sometimes an all-capitalized look is suitable for a table. By using
        `opt_all_caps()`, we can transform characters in the column labels,
        the stub, and in all row groups in this way (and there's control over which
        of these locations are transformed). This function serves as a convenient
        shortcut for `tab_options(<location>.text_transform="uppercase",
        <location>.font.size=gt.pct(80), <location>.font.weight="bolder")` (for
        all `locations` selected).

        Parameters
        ----------

        all_caps (bool):
            Indicates whether the text transformation to all caps should be
            performed (`True`, the default) or reset to default values (`False`)
            for the `locations` targeted.

        locations (list(str)):
            Which locations should undergo this text transformation? By default it
            includes all of the `"column_labels"`, the `"stub"`, and the
            `"row_group"` locations. However, we could just choose one or two
            of those.

        Returns
        -------

        GT
            Result of the table operation.
        """

        # If providing a scalar string value, normalize it to be in a list
        if type(locations).__name__ != "list":
            locations = _utils._str_scalar_to_list(cast(str, locations))

        # Ensure that the `locations` value is a list of strings
        _utils._assert_str_list(locations)

        # TODO: Ensure that all values within `locations` are valid

        # Set new options for `locations` selected, or, reset to default options
        # if `all_caps` is False
        if all_caps is True:
            if "column_labels" in locations:
                self.tab_options(column_labels_font_size="80%")
                self.tab_options(column_labels_font_weight="bolder")
                self.tab_options(column_labels_text_transform="uppercase")

            if "stub" in locations:
                self.tab_options(stub_font_size="80%")
                self.tab_options(stub_font_weight="bolder")
                self.tab_options(stub_text_transform="uppercase")

            if "row_group" in locations:
                self.tab_options(row_group_font_size="80%")
                self.tab_options(row_group_font_weight="bolder")
                self.tab_options(row_group_text_transform="uppercase")

        else:
            self.tab_options(column_labels_font_size="100%")
            self.tab_options(column_labels_font_weight="normal")
            self.tab_options(column_labels_text_transform="inherit")
            self.tab_options(stub_font_size="100%")
            self.tab_options(stub_font_weight="initial")
            self.tab_options(stub_text_transform="inherit")
            self.tab_options(row_group_font_size="100%")
            self.tab_options(row_group_font_weight="initial")
            self.tab_options(row_group_text_transform="inherit")

        return self

    # TODO: create the `opt_table_lines()` function

    # TODO: create the `opt_table_outline()` function

    # TODO: create the `opt_table_font()` function

    # TODO: create the `opt_css()` function



from __future__ import annotations

from typing import Any, List, cast
from typing_extensions import Self
import pkg_resources

import sass
import re
import copy

from great_tables._gt_data import GTData

# Main gt imports ----
from great_tables import _utils

# Rewrite main gt imports to use relative imports of APIs ----
from great_tables._tbl_data import TblDataAPI
from great_tables._body import body_reassemble
from great_tables._boxhead import BoxheadAPI
from great_tables._footnotes import FootnotesAPI
from great_tables._formats import (
    FormatsAPI,
    fmt_number,
    fmt_percent,
    fmt_integer,
    fmt_scientific,
    fmt_currency,
    fmt_engineering,
)
from great_tables._heading import HeadingAPI
from great_tables._locale import LocaleAPI
from great_tables._options import OptionsAPI
from great_tables._row_groups import RowGroupsAPI
from great_tables._source_notes import tab_source_note
from great_tables._spanners import tab_spanner, cols_move
from great_tables._stub import reorder_stub_df
from great_tables._stubhead import StubheadAPI


# from ._helpers import random_id
# from ._body import Body
from ._text import StringBuilder, _process_text
from ._utils import _as_css_font_family_attr, _unique_set
from ._tbl_data import n_rows, _get_cell


# from ._body import Context


__all__ = ["GT"]

# Architecture of GT:
# 1. GT class for holding all user specified directives (internally
#    implemented using multiple smaller modules)
# 2. Build step that performs mostly-context-agnostic pre-rendering tasks.
#    State from GT class is transformed into input for the render step.
# 3. Render into final output.


# =============================================================================
# GT class
# =============================================================================
class GT(
    GTData,
    TblDataAPI,
    BoxheadAPI,
    RowGroupsAPI,
    HeadingAPI,
    StubheadAPI,
    FootnotesAPI,
    LocaleAPI,
    FormatsAPI,
    OptionsAPI,
):
    """
    Create a gt Table object.

    Methods
    -------
        render: Renders and returns the HTML table.

    Examples
    --------
        >>> from gt import *
        >>> x = GT([{"a": 5, "b": 10}, {"a": 15, "b": 20}])
        >>> x
        >>> print(x)
    """

    def _repr_html_(self):
        return self.render(context="html")

    def __init__(
        self,
        data: Any,
        locale: str = "",
        rowname_col: str | None = None,
        groupname_col: str | None = None,
    ):
        # This is a bad idea ----
        gtdata = GTData.from_data(
            data, locale=locale, rowname_col=rowname_col, groupname_col=groupname_col
        )
        super().__init__(**gtdata.__dict__)

    # TODO: Refactor API methods -----
    fmt_number = fmt_number
    fmt_percent = fmt_percent
    fmt_integer = fmt_integer
    fmt_scientific = fmt_scientific
    fmt_currency = fmt_currency
    fmt_engineering = fmt_engineering

    tab_spanner = tab_spanner
    tab_source_note = tab_source_note
    cols_move = cols_move

    # -----

    def _get_has_built(self: GT) -> bool:
        return self._has_built

    def _render_formats(self, context: str) -> GT:
        self._body.render_formats(self._tbl_data, self._formats, context)
        return self

    def _build_data(self, context: str):
        # Build the body of the table by generating a dictionary
        # of lists with cells initially set to nan values
        built = copy.copy(self)
        built._body = self._body.__class__(self._tbl_data.copy())
        built._render_formats(context)
        # built._body = _migrate_unformatted_to_output(body)

        # built._perform_col_merge()
        built._body = body_reassemble(built._body, self._row_groups, self._stub, self._boxhead)

        # Reordering of the metadata elements of the table

        built._stub = reorder_stub_df(self._stub, self._row_groups)
        # self = self.reorder_footnotes()
        # self = self.reorder_styles()

        # Transformations of individual cells at supported locations

        # self = self.perform_text_transforms()

        # ...

        return built

    def render(self, context: str) -> str:
        self = self._build_data(context=context)

        self._has_built = True

        html_table = self._render_as_html()
        return html_table

    # =============================================================================
    # Building
    # =============================================================================

    # def _body_build(self, data: Table):
    #     return self
    #     # data.cells[(1, 3)].set_cell_value("foo")

    # def _migrate_unformatted_to_output(self, body: Dict[Column, Any]):

    #     # Get the dictionary keys from the body as these serve as column names
    #     colnames = body.keys()

    #     for column in colnames:
    #         body[column]

    #     return body

    # =============================================================================
    # HTML Rendering
    # =============================================================================
    def _render_as_html(self) -> str:
        heading_component = _create_heading_component(self)
        column_labels_component = _create_column_labels_component(self)
        body_component = _create_body_component(self)
        source_notes_component = _create_source_notes_component(self)
        footnotes_component = _create_footnotes_component(self)

        html_table = f"""<table class=\"gt_table\">
{heading_component.make_string()}
{column_labels_component}
{body_component}
{source_notes_component}
{footnotes_component}
</table>
"""

        # Get a string of compiled CSS
        css = _compile_scss(data=self)

        # Obtain the `table_id` value (might be set, might be None)
        table_id = self._options._options["table_id"].value

        # Obtain options set for overflow and container dimensions
        overflow_x = self._options._options["container_overflow_x"].value
        overflow_y = self._options._options["container_overflow_y"].value
        width = self._options._options["container_width"].value
        height = self._options._options["container_height"].value

        if table_id is None:
            id_attr_str = ""
        else:
            id_attr_str = f'id="{table_id}"'

        finalized_table = f"""<div {id_attr_str} style="overflow-x:{overflow_x};overflow-y:{overflow_y};width:{width};height:{height};">
<style>
{css}
</style>
{html_table}
</div>
        """

        return finalized_table


# =============================================================================
# End of GT class
# =============================================================================


# =============================================================================
# GT Getter/Setter Functions
# =============================================================================


def _set_has_built(gt: GT, value: bool) -> GT:
    gt._has_built = value
    return gt


def _get_column_labels(gt: GT, context: str) -> List[str]:
    gt_built = gt._build_data(context=context)
    column_labels = [x.column_label for x in gt_built._boxhead]
    return column_labels


def _get_column_of_values(gt: GT, column_name: str, context: str) -> List[str]:
    gt_built = gt._build_data(context=context)
    tbl_data = gt_built._body.body
    cell_values: List[str] = []

    for i in range(n_rows(tbl_data)):
        cell_content: Any = _get_cell(tbl_data, i, column_name)
        cell_str: str = str(cell_content)
        cell_values.append(cell_str)

    return cell_values


# =============================================================================
# Table Structuring Functions
# =============================================================================


def _create_heading_component(data: GT) -> StringBuilder:
    result = StringBuilder()

    title = data._heading.title
    subtitle = data._heading.subtitle

    has_title = _utils.heading_has_title(title=title)
    has_subtitle = _utils.heading_has_subtitle(subtitle=subtitle)

    # If there is no title or heading component, then return an empty string
    if not has_title and not has_subtitle:
        return result

    title = _process_text(title)
    subtitle = _process_text(subtitle)

    # Get the effective number of columns, which is number of columns
    # that will finally be rendered accounting for the stub layout
    n_cols_total = data._boxhead._get_effective_number_of_columns()

    result.append(
        f"""  <tr>
    <th colspan="{n_cols_total}" class="gt_heading gt_title gt_font_normal">{title}
  </tr>"""
    )

    if has_subtitle:
        subtitle_row = f"""  <tr>
    <th colspan="{n_cols_total}" class="gt_heading gt_subtitle gt_font_normal gt_bottom_border">{subtitle}
  </tr>"""
        result.append(f"\n{subtitle_row}")

    return StringBuilder('<thead class="gt_header">', result, "</thead>")


def _create_column_labels_component(data: GT) -> str:
    column_names = data._boxhead._get_column_labels()

    th_cells = "".join([f"  <th>{x}</th>\n" for x in column_names])

    column_names_str = f"<tr>\n{th_cells}</tr>"

    return column_names_str


def _create_body_component(data: GT):
    import pandas as pd

    # for now, just coerce everything in the original data to a string
    # so we can fill in the body data with it
    _str_orig_data = data._tbl_data.applymap(lambda x: str(x) if not pd.isna(x) else x)

    tbl_data = data._body.body.fillna(_str_orig_data)

    column_names = data._boxhead._get_visible_columns()

    body_rows: List[str] = []

    for i in range(n_rows(tbl_data)):
        body_cells: List[str] = []

        for name in column_names:
            cell_content: Any = _get_cell(tbl_data, i, name)
            cell_str: str = str(cell_content)

            body_cells.append('  <td class="gt_row">' + cell_str + "</td>")

        body_rows.append("<tr>\n" + "\n".join(body_cells) + "\n</tr>")

    all_body_rows = "\n".join(body_rows)

    return f'<tbody class="gt_table_body">\n{all_body_rows}\n</tbody>'


def _get_stub_layout(data: GT) -> List[str]:
    # Determine which stub components are potentially present as columns
    stub_rownames_is_column = _stub_rownames_has_column(data=data)
    stub_groupnames_is_column = _stub_group_names_has_column(data=data)

    # Get the potential total number of columns in the table stub
    n_stub_cols = stub_rownames_is_column + stub_groupnames_is_column

    # Resolve the layout of the stub (i.e., the roles of columns if present)
    if n_stub_cols == 0:
        # If summary rows are present, we will use the `rowname` column
        # for the summary row labels
        if _summary_exists(data=data):
            stub_layout = ["rowname"]
        else:
            stub_layout = []

    else:
        stub_layout = [
            label
            for label, condition in [
                ("group_label", stub_groupnames_is_column),
                ("rowname", stub_rownames_is_column),
            ]
            if condition
        ]

    return stub_layout


# Determine whether the table should have row labels set within a column in the stub
def _stub_rownames_has_column(data: GT) -> bool:
    return "row_id" in _get_stub_components(data=data)


# Determine whether the table has any row labels or row groups defined and provide
# a simple list that contains at a maximum two components
def _get_stub_components(data: GT):
    # TODO: we should be using `row_id` instead of `rowname`
    # Obtain the object that describes the table stub
    tbl_stub = data._stub

    # Get separate lists of `group_id` and `row_id` values from the `_stub` object
    group_id_vals = [tbl_stub[i].group_id for i in range(len(tbl_stub))]
    rowname_vals = [tbl_stub[i].rowname for i in range(len(tbl_stub))]

    stub_components: list[str] = []

    if any(x is not None for x in group_id_vals):
        stub_components.append("group_id")

    if any(x is not None for x in rowname_vals):
        stub_components.append("row_id")

    return stub_components


# Determine whether the table should have row group labels set within a column in the stub
def _stub_group_names_has_column(data: GT) -> bool:
    # If there aren't any row groups then the result is always False
    if len(_row_groups_get(data=data)) < 1:
        return False

    # Given that there are row groups, we need to look at the option `row_group_as_column` to
    # determine whether they populate a column located in the stub; if set as True then that's
    # the return value
    row_group_as_column = data._options._get_option_value(option="row_group_as_column")

    row_group_as_column: Any
    if not isinstance(row_group_as_column, bool):
        raise TypeError("Variable type mismatch. Expected bool, got something entirely different.")

    return row_group_as_column


def _row_groups_get(data: GT) -> List[str]:
    return data._row_groups._d


def _create_source_notes_component(data: GT) -> str:
    source_notes = data._source_notes

    # If there are no source notes, then return an empty string
    if source_notes == []:
        return ""

    # Obtain the `multiline` and `separator` options from `_options`
    multiline = data._options.source_notes_multiline.value
    separator = cast(str, data._options.source_notes_sep.value)

    # Get the effective number of columns, which is number of columns
    # that will finally be rendered accounting for the stub layout
    n_cols_total = data._boxhead._get_effective_number_of_columns()

    # Handle the multiline source notes case (each note takes up one line)
    if multiline:
        # Create the source notes component as a series of `<tr><td>` (one per
        # source note) inside of a `<tfoot>`

        source_notes_tr: List[str] = []

        for note in source_notes:
            source_notes_tr.append(
                f"""
  <tr>
    <td class="gt_sourcenote" colspan="{n_cols_total}">{note}</td>
  </tr>
"""
            )

        source_notes_joined = "\n".join(source_notes_tr)

        source_notes_component = f"""  <tfoot class="gt_sourcenotes">
  {source_notes_joined}
</tfoot>"""

        return source_notes_component

    # TODO: Perform HTML escaping on the separator text and
    # transform space characters to non-breaking spaces

    # Create the source notes component as a single `<tr><td>` inside
    # of a `<tfoot>`

    source_notes_str_joined = separator.join(source_notes)

    source_notes_component = f"""<tfoot>
  <tr class="gt_sourcenotes">
    <td class="gt_sourcenote" colspan="{n_cols_total}">
      <div style="padding-bottom:2px;">{source_notes_str_joined}</div>
    </td>
  </tr>
</tfoot>
    """

    return source_notes_component


def _create_footnotes_component(data: GT):
    return ""


# TODO: Port the SCSS compilation routine from the R implementation here
def _compile_scss(data: GT) -> str:
    # Obtain the SCSS options dictionary
    gt_options_dict = data._options._options

    # Get collection of parameters that pertain to SCSS
    scss_params = [
        f"${x.parameter}: {x.value};"
        for x in gt_options_dict.values()
        if x.scss is True and x.value is not None
    ]
    scss_params_str = "\n".join(scss_params) + "\n"

    # Obtain the `table_id` value (might be set, might be None)
    table_id = gt_options_dict["table_id"].value

    # TODO: need to implement a function to normalize color (`html_color()`)

    # Get the unique list of fonts from `gt_options_dict`
    font_list = _unique_set(gt_options_dict["table_font_names"].value)

    # Generate a `font-family` string
    if font_list is not None:
        font_family_attr = _as_css_font_family_attr(fonts=font_list)
    else:
        font_family_attr = ""

    gt_styles_default_file = open(
        pkg_resources.resource_filename("great_tables", "css/gt_styles_default.scss")
    )

    gt_styles_default = gt_styles_default_file.read()
    gt_styles_default = re.sub(r"\s+", " ", gt_styles_default, 0, re.MULTILINE)
    gt_styles_default = re.sub(r"}", "}\n", gt_styles_default, 0, re.MULTILINE)

    gt_colors_file = open(pkg_resources.resource_filename("great_tables", "css/gt_colors.scss"))

    gt_colors = gt_colors_file.read()

    scss = scss_params_str + gt_colors + gt_styles_default

    compiled_css = cast(str, sass.compile(string=scss))
    if table_id is not None:
        compiled_css = re.sub(r"\.gt_", f"#{table_id} .gt_", compiled_css, 0, re.MULTILINE)

    finalized_css = f"""html {{
      {font_family_attr}
}}

{compiled_css}"""

    return finalized_css



from typing import Union, List
import random

from ._text import Text


def letters() -> List[str]:
    """Lowercase letters of the Roman alphabet

    Returns:
        List[str]: the 26 lowercase letters of the Roman alphabet
    """
    lett = [
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
        "m",
        "n",
        "o",
        "p",
        "q",
        "r",
        "s",
        "t",
        "u",
        "v",
        "w",
        "x",
        "y",
        "z",
    ]

    return lett


def LETTERS() -> List[str]:
    """Uppercase letters of the Roman alphabet

    Returns:
        List[str]: the 26 uppercase letters of the Roman alphabet
    """
    lett = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
        "K",
        "L",
        "M",
        "N",
        "O",
        "P",
        "Q",
        "R",
        "S",
        "T",
        "U",
        "V",
        "W",
        "X",
        "Y",
        "Z",
    ]

    return lett


def px(x: Union[int, float]) -> str:
    """
    Helper for providing a CSS length value in pixels.

    For certain parameters, a length value is required. Examples include the
    setting of font sizes (e.g., in `cell_text()`) and thicknesses of lines
    (e.g., in `cell_borders()`). Setting a length in pixels with `px()` allows
    for an absolute definition of size as opposed to the analogous helper
    function `pct()`.

    Parameters
    ----------
    x (Union[int, float])
        The integer or float value to format as a string (e.g., `"12px"`) for
        some `tab_options()` arguments that can take values as units of
        pixels (e.g., `table_font_size`).

    Examples
    --------
        >>> from gt import *
        >>> x = gt.px(12)
        >>> x
        >>> print(x)
    """
    return f"{x}px"


def pct(x: Union[int, float]) -> str:
    """
    Helper for providing a CSS length value as a percentage.

    A percentage value acts as a length value that is relative to an initial
    state. For instance an 80% value for something will size the target
    to 80% the size of its 'previous' value. This type of sizing is
    useful for sizing up or down a length value with an intuitive measure. This
    helper function can be used for the setting of font sizes (e.g., in
    `cell_text()`) and altering the thicknesses of lines (e.g., in
    `cell_borders()`. Should a more exact definition of size be required, the
    analogous helper function `pct()` will be more useful.

    Parameters
    ----------
    x (Union[int, float])
        The integer or float value to format as a string-based percentage value
        for some `tab_options()` arguments that can take percentage values
        (e.g., `table.width`).

    Examples
    --------
        >>> from gt import *
        >>> x = gt.pct(80)
        >>> x
        >>> print(x)
    """
    return f"{x}%"


def md(text: str) -> Text:
    """Interpret input text as Markdown-formatted text.

    Markdown! It's a wonderful thing. We can use it in certain places (e.g.,
    footnotes, source notes, the table title, etc.) and expect it to render to
    HTML as Markdown does. There is the `html()` helper that allows you to ferry
    in HTML but this function `md()`... it's almost like a two-for-one deal (you
    get to use Markdown plus any HTML fragments at the same time).

    Args:
        text (str): The text that is understood to contain Markdown
        formatting.

    Returns:
        Text: An instance of the Text class is returned, where the text
        `type` is 'markdown'.
    """
    return Text(text=text, type="from_markdown")


def html(text: str) -> Text:
    """Interpret input text as HTML-formatted text.

    For certain pieces of text (like in column labels or table headings) we may
    want to express them as raw HTML. In fact, with HTML, anything goes so it
    can be much more than just text. The `html()` function will guard the input
    HTML against escaping, so, your HTML tags will come through as HTML when
    rendered... to HTML.

    Args:
        text (str): The text that is understood to contain HTML formatting.

    Returns:
        Text: An instance of the Text class is returned, where the text
        `type` is 'html'.
    """
    return Text(text=text, type="html")


def random_id(n: int = 10) -> str:
    """Helper for creating a random `id` for a gt table

    Args:
        n (int, optional): The number of lowercase letters to use in the random
        ID string. Defaults to 10.

    Returns:
        str: A string that constitutes a random ID value.
    """
    return "".join(random.choices(letters(), k=n))



from __future__ import annotations

from typing import TYPE_CHECKING
from .utils_render_common import get_row_reorder_df
from ._tbl_data import reorder


if TYPE_CHECKING:
    from ._gt_data import Body, RowGroups, Stub, Boxhead


def body_reassemble(body: Body, row_groups: RowGroups, stub_df: Stub, boxhead: Boxhead) -> Body:
    cols = [col_info.var for col_info in boxhead]

    start_final = get_row_reorder_df(row_groups, stub_df)
    rows = [final for _, final in start_final]

    # TODO: once body is just a DataFrame, we can call reorder directly on it
    return body.__class__(reorder(body.body, rows, cols))



from __future__ import annotations
from typing import Optional, Union, List


class HeadingAPI:
    def tab_header(
        self,
        title: str,
        subtitle: Optional[str] = None,
        preheader: Optional[Union[str, List[str]]] = None,
    ):
        """
        Add a table header.

        We can add a table header to the gt table with a title and even a
        subtitle. A table header is an optional table part that is positioned
        above the column labels. We have the flexibility to use Markdown
        formatting for the header's title and subtitle. Furthermore, if the
        table is intended for HTML output, we can use HTML in either of the
        title or subtitle.

        Parameters
        ----------
        title (str), subtitle (str)
            Text to be used in the table title and, optionally, for
            the table subtitle. We can elect to use the `md()` and `html()`
            helper functions to style the text as Markdown or to retain HTML
            elements in the text.
        preheader (str)
            Optional preheader content that is rendered above the table. Can
            be supplied as a list of strings.

        Returns
        -------
        GT
            Result of the table operation.

        Examples
        --------
            >>> from gt import *
            >>> x = GT([{"a": 5, "b": 10}, {"a": 15, "b": 20}])
            >>>     .tab_header(title="Title", subtitle="Subtitle")
            >>> x
            >>> print(x)
        """
        self._heading.title = title
        self._heading.subtitle = subtitle
        self._heading.preheader = preheader

        return self



from __future__ import annotations

from ._gt_data import GTData


def tab_source_note(data: GTData, source_note: str):
    """
    Add a source note citation.

    Add a source note to the footer part of the gt table. A source note is
    useful for citing the data included in the table. Several can be added to the
    footer, simply use multiple calls of `tab_source_note()` and they will be
    inserted in the order provided. We can use Markdown formatting for the note,
    or, if the table is intended for HTML output, we can include HTML formatting.

    Parameters
    ----------
    source_note (str)
        Text to be used in the source note. We can optionally use the `md()` or
        `html()` helper functions to style the text as Markdown or to retain HTML
        elements in the text.

    Returns
    -------
    GT
        Result of the table operation.

    Examples
    --------
        >>> from gt import *
        >>> x = GT([{"a": 5, "b": 10}, {"a": 15, "b": 20}])
        >>>     .tab_source_note(source_note="Source note for the table.")
        >>> x
        >>> print(x)
    """

    return data._replace(_source_notes=data._source_notes + [source_note])



from great_tables import *
import pandas as pd


def test_render_doc_module(snapshot):
    pd_data = pd.DataFrame([{"a": 5.23, "b": 15}, {"a": 15, "b": 2000}])

    res = gt.GT(pd_data)._repr_html_()

    assert res == snapshot



import pandas as pd
import polars as pl
import polars.testing
import pytest

from great_tables._tbl_data import (
    _get_cell,
    _get_column_dtype,
    _set_cell,
    get_column_names,
    DataFrameLike,
    reorder,
    eval_select,
)


params_frames = [pytest.param(pd.DataFrame, id="pandas"), pytest.param(pl.DataFrame, id="polars")]


@pytest.fixture(params=params_frames, scope="function")
def df(request) -> pd.DataFrame:
    return request.param({"col1": [1, 2, 3], "col2": ["a", "b", "c"], "col3": [4.0, 5.0, 6.0]})


def assert_frame_equal(src, target):
    if isinstance(src, pd.DataFrame):
        pd.testing.assert_frame_equal(src, target)
    elif isinstance(src, pl.DataFrame):
        pl.testing.assert_frame_equal(src, target)
    else:
        raise NotImplementedError(f"Unsupported data type: {type(src)}")


def test_get_column_names(df: DataFrameLike):
    expected = ["col1", "col2", "col3"]
    assert get_column_names(df) == expected


def test_get_column_dtypes(df: DataFrameLike):
    assert _get_column_dtype(df, "col1") == df["col1"].dtype


def test_get_cell(df: DataFrameLike):
    assert _get_cell(df, 1, "col2") == "b"


def test_set_cell(df: DataFrameLike):
    expected = df.__class__({"col1": [1, 2, 3], "col2": ["a", "x", "c"], "col3": [4.0, 5.0, 6.0]})
    _set_cell(df, 1, "col2", "x")
    assert_frame_equal(df, expected)


def test_reorder(df: DataFrameLike):
    res = reorder(df, [0, 2], ["col2"])
    dst = df.__class__({"col2": ["a", "c"]})

    if isinstance(dst, pd.DataFrame):
        dst.index = pd.Index([0, 2])

    assert_frame_equal(res, dst)


def test_eval_select_with_list(df: DataFrameLike):
    sel = eval_select(df, ["col2", "col1"])
    assert sel == [("col2", 1), ("col1", 0)]



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple

if TYPE_CHECKING:
    from ._gt_data import RowGroups, Stub


TupleStartFinal = Tuple[int, int]


def get_row_reorder_df(groups: RowGroups, stub_df: Stub) -> list[TupleStartFinal]:
    # Get the number of non-None entries in the `groupname_col`
    n_stub_entries = len([entry for entry in stub_df if entry.group_id is not None])

    # Raise a ValueError if there are row group entries but no RowGroups
    if n_stub_entries and not len(groups):
        raise ValueError(f"Detected {n_stub_entries} but only {len(groups)} row groups.")

    # If there aren't any row groups then return a list of tuples that don't lead
    # to any resorting later on (e.g., `[(0, 0), (1, 1), (2, 2) ... (n, n)]`)
    if not len(groups):
        indices = range(len(stub_df))

        # TODO: is this used in indexing? If so, we may need to use
        # ii + 1 for the final part?
        return [(ii, ii) for ii in indices]

    # where in the group each element is
    # TODO: this doesn't yield consistent values
    groups_pos = [groups.index(row.group_id) if row.group_id is not None else -1 for row in stub_df]

    # From running test_body_reassemble():
    # print(groups_pos)
    # [0, 1, 0, 1] <- correct
    # [1, 0, 1, 0] <- wrong

    # the index that when used on the rows will sort them by the order in groups
    start_pos = list(range(len(groups_pos)))
    sort_indx = sorted(start_pos, key=lambda ii: groups_pos[ii])

    # From running test_body_reassemble():
    # [(0, 0), (1, 2), (2, 1), (3, 3)] <- correct
    # [(0, 1), (1, 3), (2, 0), (3, 2)] <- wrong
    return list(zip(start_pos, sort_indx))



import pandas as pd
import pytest

from great_tables._spanners import (
    spanners_print_matrix,
    empty_spanner_matrix,
    tab_spanner,
    cols_move,
)
from great_tables._gt_data import Spanners, SpannerInfo, Boxhead, ColInfo, ColInfoTypeEnum
from great_tables import GT


@pytest.fixture
def spanners():
    return Spanners(
        [
            SpannerInfo(spanner_id="a", spanner_level=0, vars=["col1"], built="A"),
            SpannerInfo(spanner_id="b", spanner_level=1, vars=["col2"], built="B"),
        ]
    )


@pytest.fixture
def boxhead():
    return Boxhead(
        [
            ColInfo(var="col1"),
            ColInfo(var="col2"),
            ColInfo(var="col3"),
            ColInfo(var="col4", type=ColInfoTypeEnum.hidden),
        ]
    )


def test_spanners_print_matrix(spanners, boxhead):
    mat, vars = spanners_print_matrix(spanners, boxhead)
    assert vars == ["col1", "col2", "col3"]
    assert mat == [
        {"col1": None, "col2": "B", "col3": None},
        {"col1": "A", "col2": None, "col3": None},
        {"col1": "col1", "col2": "col2", "col3": "col3"},
    ]


def test_spanners_print_matrix_arg_omit_columns_row(spanners, boxhead):
    mat, vars = spanners_print_matrix(spanners, boxhead, omit_columns_row=True)
    assert vars == ["col1", "col2", "col3"]
    assert mat == [
        {"col1": None, "col2": "B", "col3": None},
        {"col1": "A", "col2": None, "col3": None},
    ]


def test_spanners_print_matrix_arg_include_hidden(spanners, boxhead):
    mat, vars = spanners_print_matrix(spanners, boxhead, include_hidden=True)
    assert vars == ["col1", "col2", "col3", "col4"]
    assert mat == [
        {"col1": None, "col2": "B", "col3": None, "col4": None},
        {"col1": "A", "col2": None, "col3": None, "col4": None},
        {"col1": "col1", "col2": "col2", "col3": "col3", "col4": "col4"},
    ]


def test_empty_spanner_matrix():
    mat, vars = empty_spanner_matrix(["a", "b"], omit_columns_row=False)

    assert vars == ["a", "b"]
    assert mat == [{"a": "a", "b": "b"}]


def test_empty_spanner_matrix_arg_omit_columns_row():
    mat, vars = empty_spanner_matrix(["a", "b"], omit_columns_row=True)

    assert vars == ["a", "b"]
    assert mat == []


def test_tab_spanners_with_columns():
    df = pd.DataFrame({"a": [1, 2], "b": [3, 4], "c": [5, 6]})
    src_gt = GT(df)

    dst_span = SpannerInfo("a_spanner", 0, "a_spanner", vars=["b", "a"])

    new_gt = tab_spanner(src_gt, "a_spanner", columns=["b", "a"], gather=False)
    assert len(new_gt._spanners)
    assert new_gt._spanners[0] == dst_span


def test_tab_spanners_with_spanner_ids():
    df = pd.DataFrame({"a": [1, 2], "b": [3, 4], "c": [5, 6]})
    src_gt = GT(df)

    # we'll be testing for this second spanner added
    dst_span = SpannerInfo("b_spanner", 1, "b_spanner", vars=["c", "b", "a"])

    gt_with_span = tab_spanner(src_gt, "a_spanner", columns=["b", "a"], gather=False)

    new_gt = tab_spanner(gt_with_span, "b_spanner", spanners="a_spanner", columns=["c"])

    assert len(new_gt._spanners) == 2
    assert new_gt._spanners[1] == dst_span


def test_tab_spanners_with_gather():
    df = pd.DataFrame({"a": [1, 2], "b": [3, 4], "c": [5, 6]})
    src_gt = GT(df)

    new_gt = tab_spanner(src_gt, "a_spanner", columns=["a", "c"], gather=True)

    assert len(new_gt._spanners) == 1
    assert [col.var for col in new_gt._boxhead] == ["a", "c", "b"]


def test_cols_move():
    df = pd.DataFrame({"a": [1, 2], "b": [3, 4], "c": [5, 6]})
    src_gt = GT(df)

    new_gt = cols_move(src_gt, columns=["a"], after="b")
    assert [col.var for col in new_gt._boxhead] == ["b", "a", "c"]



from __future__ import annotations

from ._gt_data import Stub, RowGroups
from .utils_render_common import get_row_reorder_df


def reorder_stub_df(stub_df: Stub, row_groups: RowGroups) -> Stub:
    start_final = get_row_reorder_df(row_groups, stub_df)

    stub_df = stub_df[[final for _, final in start_final]]

    return stub_df



import pytest

from great_tables import GT
from great_tables._gt_data import RowGroups
import pandas as pd

# Generate a gt Table object for assertion testing
data = [{"a": 5, "b": 15}, {"a": 15, "b": 2000}]
pd_data = pd.DataFrame(data)


@pytest.fixture
def gt_tbl():
    return GT(pd_data)


def test_gt_replace(gt_tbl: GT):
    row_groups = RowGroups(["x"])
    new_gt_tbl = gt_tbl._replace(_row_groups=row_groups)

    assert new_gt_tbl._row_groups is row_groups


def test_gt_object_prerender(gt_tbl: GT):
    assert type(gt_tbl).__name__ == "GT"

    # Assert that certain private variables exist with a
    # GT object and that they have the expected names
    assert type(gt_tbl._boxhead).__name__ == "Boxhead"
    assert type(gt_tbl._stub).__name__ == "Stub"
    assert type(gt_tbl._row_groups).__name__ == "RowGroups"
    assert type(gt_tbl._spanners).__name__ == "Spanners"
    assert type(gt_tbl._heading).__name__ == "Heading"
    assert type(gt_tbl._stubhead).__name__ == "Stubhead"
    assert isinstance(gt_tbl._source_notes, list)
    assert type(gt_tbl._footnotes).__name__ == "Footnotes"
    assert type(gt_tbl._styles).__name__ == "Styles"
    assert type(gt_tbl._locale).__name__ == "Locale"


def test_gt_table_render(gt_tbl: GT):
    # Assert that a table render process will generate a string object
    assert type(gt_tbl.render(context="html")).__name__ == "str"

    assert (
        type(
            gt_tbl.tab_header(title="Title of Table", subtitle="The Subtitle").render(
                context="html"
            )
        ).__name__
        == "str"
    )

    assert type(gt_tbl.tab_source_note(source_note="Note").render(context="html")).__name__ == "str"

    assert (
        type(
            gt_tbl.tab_source_note(source_note="Note 1")
            .tab_source_note(source_note="Note 2")
            .render(context="html")
        ).__name__
        == "str"
    )



from typing import Union, List

import commonmark
import re


class Text:
    def __init__(self, text: str, type: str):
        self.text: str = text
        self.type: str = type


class StringBuilder:
    pieces: List[Union[str, "StringBuilder"]]

    def __init__(self, *args: Union[str, "StringBuilder"]):
        self.pieces = list(args)

    def _collect(self, lst: List[str]):
        for piece in self.pieces:
            if isinstance(piece, str):
                lst.append(piece)
            else:
                piece._collect(lst)

    def make_string(self) -> str:
        lst = []
        self._collect(lst)
        return "".join(lst)

    def append(self, *args: str):
        self.pieces.extend(args)

    def prepend(self, *args: str):
        self.pieces[0:0] = args


def _md_html(x: str) -> str:
    str = commonmark.commonmark(x)
    return re.sub(r"^<p>|</p>\n$", "", str)


def _process_text(x: Union[Text, str, None]) -> str:
    if x is None:
        return ""

    if isinstance(x, str):
        text = x
        type = "plaintext"
    else:
        text = x.text
        type = x.type

    if type == "from_markdown":
        x_out = _md_html(text)
    elif type == "html":
        x_out = text
    else:
        # TODO: Perform HTML escaping
        x_out = text

    return x_out



from __future__ import annotations

from ._gt_data import GTData
from ._locations import Loc

from dataclasses import dataclass
from typing import Optional, Literal


# Cell Styles ==========================================================================
# TODO: stubbed out the styles in helpers.R as dataclasses while I was reading it,
# but have no worked on any runtime validation, etc..


@dataclass
class CellStyle:
    """A style specification."""


@dataclass
class CellText:
    """A style specification for text."""

    color: str
    font: str
    size: str
    align: Literal["center", "left", "right", "justify"]
    # TODO: this can also be a gt_column object?
    v_align: Literal["middle", "top", "bottom"]
    style: Literal["normal", "italic", "oblique"]
    weight: Literal["normal", "bold", "bolder", "lighter"]
    stretch: Literal[
        "normal",
        "condensed",
        "ultra-condensed",
        "extra-condensed",
        "semi-condensed",
        "semi-expanded",
        "expanded",
        "extra-expanded",
        "ultra-expanded",
    ]
    decorate: Literal["overline", "line-through", "underline", "underline overline"]
    transform: Literal["uppercase", "lowercase", "capitalize"]
    whitespace: Literal["normal", "nowrap", "pre", "pre-wrap", "pre-line", "break-spaces"]


@dataclass
class CellStyleFill(CellStyle):
    """A style specification for fill."""

    fill: str
    alpha: Optional[float] = None


@dataclass
class CellStyleBorders(CellStyle):
    sides: Literal["all", "top", "bottom", "left", "right"]
    color: str
    style: str
    # TODO: this can include objects like px(1)
    weight: str


def tab_style(data: GTData, style: list[CellStyle] | CellStyle, locations: Loc):
    """Add custom style to one or more cells

    Parameters
    ----------
    style:
        A style specification.
    location:
        A location on the table.
    """

    raise NotImplementedError()



# Set version ----
from importlib_metadata import version as _v

__version__ = _v("great_tables")

del _v

# Main gt imports ----

from .gt import GT
from great_tables.data import *

# from ._base_api import *
# from ._body import *
# from ._boxhead import *
# from ._footnotes import *
# from ._formats import *
# from ._heading import *
from ._helpers import letters, LETTERS, px, pct, md, html, random_id
# from ._locale import *
# from ._options import *
# from ._row_groups import *
# from ._source_notes import *
# from ._spanners import *
# from ._stub import *
# from ._stubhead import *
# from ._styles import *
# from ._table import *
# from ._tbl_data import *
# from ._text import *
# from ._utils import *

#



from typing import Optional, Union, List, Any
import pandas as pd
import json


def heading_has_title(title: Optional[str]) -> bool:
    return title is not None


def heading_has_subtitle(subtitle: Optional[str]) -> bool:
    return subtitle is not None


def _match_arg(x: str, lst: List[str]) -> str:
    # Ensure that `lst` has at least one element
    if len(lst) == 0:
        raise ValueError("The `lst` object must contain at least one element.")

    # Ensure that all elements in `lst` are strings
    if not all(isinstance(el, str) for el in lst):
        raise ValueError("All elements in the `lst` object must be strings.")

    # Ensure that `lst` does not have duplicates by comparing against
    # a set based on `lst` (using `set()` will remove duplicates)
    if len(lst) != len(set(lst)):
        raise ValueError("The `lst` object must contain unique elements.")

    matched = [el for el in lst if x in el]

    # Raise error if there is no match
    if len(matched) == 0:
        raise ValueError(f"The supplied value (`{x}`) is not an allowed option.")

    return matched.pop()


def _assert_str_scalar(x: Any) -> None:
    if type(x).__name__ != "str":
        raise AssertionError(f"The supplied value (`{x}`) is not a string.")


def _assert_str_list(x: Any) -> None:
    if type(x).__name__ != "list":
        raise AssertionError(f"The supplied value (`{x}`) is not a list.")
    if not all(map(lambda x: isinstance(x, str), x)):
        raise AssertionError("Not all elements of the supplied list are strings.")


def _assert_str_in_set(x: str, set: List[str]):
    while x not in set:
        raise AssertionError(f"The string `{x}` is not part of the defined `set`.")


def _assert_list_is_subset(x: List[Any], set: List[Any]):
    if not (all(x in x for x in set)):
        raise AssertionError("The `x` list is not a strict subset of the defined `set`.")


def _str_scalar_to_list(x: str):
    _assert_str_scalar(x)
    return [x]


def _unique_set(x: Union[List[Any], None]) -> Union[List[Any], None]:
    if x is None:
        return None
    return list({k: True for k in x})


def _as_css_font_family_attr(fonts: List[str], value_only: bool = False) -> str:
    fonts_w_spaces = list(map(lambda x: f"'{x}'" if " " in x else x, fonts))

    fonts_str = ", ".join(fonts_w_spaces)

    if value_only is True:
        return fonts_str

    return f"font-family: {fonts_str};"


def _object_as_dict(v: Any) -> Any:
    try:
        return v.object_as_dict()
    except:
        pass
    if type(v) == pd.DataFrame:
        return v.to_dict()
    if type(v) in [tuple, list]:
        return list(_object_as_dict(i) for i in v)
    if type(v) == dict:
        return dict((k, _object_as_dict(val)) for (k, val) in v.items())
    if type(v) == type(_object_as_dict):  # FIXME figure out how to get "function"
        return f"<function {v.__name__}>"
    try:
        d = vars(v)
    except TypeError:
        try:
            json.dumps(v)
        except TypeError:
            return "JSON_UNSERIALIZABLE"
        return v
    return dict((k, _object_as_dict(v)) for (k, v) in d.items())


def prettify_gt_object(v: Any) -> str:
    return json.dumps(_object_as_dict(v), indent=2)



from __future__ import annotations

import pandas as pd

from ._tbl_data import get_column_names

from ._utils import _assert_list_is_subset


class BoxheadAPI():
    def cols_label(self, **kwargs: str):
        """
        Relabel one or more columns.

        Column labels can be modified from their default values (the names of
        the columns from the input table data). When you create a gt table
        object using `gt.GT()`, column names effectively become the column
        labels. While this serves as a good first approximation, column names
        aren't often appealing as column labels in a gt output table. The
        `cols_label()` method provides the flexibility to relabel one or more
        columns and we even have the option to use the `md()` or `html()`
        helpers for rendering column labels from Markdown or using HTML.

        It's important to note that while columns can be freely relabeled, we
        continue to refer to columns by their original column names. Column
        names in a data frame must be unique whereas column labels in gt have
        no requirement for uniqueness (which is useful for labeling columns as,
        say, measurement units that may be repeated several times---usually
        under different spanner column labels). Thus, we can still easily
        distinguish between columns in other gt method calls (e.g., in all of
        the `fmt*()` methods) even though we may lose distinguishability in
        column labels once they have been relabeled.

        Returns
        -------
        GT
            Result of the table operation.
        """

        # If nothing is provided, return `data` unchanged
        if len(kwargs) == 0:
            return self

        mod_columns = list(kwargs.keys())
        new_labels = list(kwargs.values())

        # Get the full list of column names for the data
        column_names = self._boxhead._get_columns()

        # Stop function if any of the column names specified are not in `cols_labels`
        # msg: "All column names provided must exist in the input `.data` table."
        _assert_list_is_subset(mod_columns, column_names)

        for i in range(len(kwargs)):
            self._boxhead._set_column_label(column=mod_columns[i], label=new_labels[i])

        return self

    def _print_boxhead(self) -> pd.DataFrame:
        boxhead_list = list(
            zip(
                [x.var for x in self._boxhead],
                [x.visible for x in self._boxhead],
                [x.column_label for x in self._boxhead],
            )
        )
        return pd.DataFrame(boxhead_list, columns=["var", "visible", "column_label"])
